# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import warnings

from webob import Request

from nose.tools import istest
from reahl.tofu import test
from reahl.tofu import vassert
from reahl.stubble import stubclass

from reahl.web.fw import UserInterface, Widget, FactoryDict, UserInterfaceFactory, RegexPath
from reahl.web.fw import Region
from reahl.web.ui import HTML5Page, P, A, Panel, Slot
from reahl.web.pure import PageColumnLayout
from reahl.webdev.tools import Browser, WidgetTester
from reahl.web_dev.fixtures import WebFixture


@istest
class UserInterfaceTests(object):
    @test(WebFixture)
    def basic_ui(self, fixture):
        """A UserInterface is a chunk of web app that can be grafted onto the URL hierarchy of any app.
        
           A UserInterface has its own views. Its Views are relative to the UserInterface itself.
        """
        class UIWithTwoViews(UserInterface):
            def assemble(self):
                self.define_view('/', title='UserInterface root view')
                self.define_view('/other', title='UserInterface other view')

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_user_interface('/a_ui',  UIWithTwoViews,  {}, name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open('/a_ui/')
        vassert( browser.title == 'UserInterface root view' )

        browser.open('/a_ui/other')
        vassert( browser.title == 'UserInterface other view' )

    @test(WebFixture)
    def backwards_compatibility(self, fixture):
        """For backwards compatibility, there are aliases for new names:
             
             - UserInterface is Region
             - .define_user_interface() is define_region()
             - .define_page() is define_main_window()
        """
        class UIWithAView(UserInterface):
            def assemble(self):
                self.define_view('/aview', title='UserInterface view')

        class MainUI(Region):
            def assemble(self):
                self.define_main_window(HTML5Page)
                self.define_region('/a_ui',  UIWithAView,  {}, name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        with warnings.catch_warnings(record=True) as caught_warnings:
            warnings.simplefilter("always")
            browser.open('/a_ui/aview')
            vassert( browser.title == 'UserInterface view' )

        vassert( caught_warnings )

    @test(WebFixture)
    def ui_slots_map_to_window(self, fixture):
        """The UserInterface uses its own names for Slots. When attaching a UserInterface, you have to specify 
            which of the UserInterface's Slots plug into which of the page's Slots.
        """
        class UIWithSlots(UserInterface):
            def assemble(self):
                root = self.define_view('/', title='UserInterface root view')
                root.set_slot('text', P.factory(text='in user_interface slot named text'))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
                self.define_user_interface('/a_ui',  UIWithSlots,  {'text': 'main'}, name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open('/a_ui/')
        vassert( browser.title == 'UserInterface root view' )

        # The widget in the UserInterface's slot named 'text' end up in the HTML5Page slot called main
        [p] = browser.lxml_html.xpath('//div[contains(@class,"column-main")]/p')
        vassert( p.text == 'in user_interface slot named text' )


    @test(WebFixture)
    def ui_redirect(self, fixture):
        """When opening an URL without trailing slash that maps to where a UserInterface is attached,
           the browser is redirected to the UserInterface '/' View."""
           
        class UIWithRootView(UserInterface):
            def assemble(self):
                self.define_view('/', title='UserInterface root view')

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_user_interface('/a_ui',  UIWithRootView,  {}, name='myui')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open('/a_ui')
        vassert( browser.title == 'UserInterface root view' )
        vassert( browser.location_path == '/a_ui/' )

    @test(WebFixture)
    def ui_arguments(self, fixture):
        """UserInterfaces can take exta args and kwargs."""
           
        class UIWithArguments(UserInterface):
            def assemble(self, kwarg=None):
                self.kwarg = kwarg
                text = self.kwarg
                root = self.define_view('/', title='A view')
                root.set_slot('text', P.factory(text=text))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
                self.define_user_interface('/a_ui', UIWithArguments, {'text': 'main'},
                                name='myui', kwarg='the kwarg')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        browser = Browser(wsgi_app)

        browser.open('/a_ui/')
        [p] = browser.lxml_html.xpath('//p')
        vassert( p.text == 'the kwarg' )

    @test(WebFixture)
    def bookmarks(self, fixture):
        """Bookmarks are pointers to Views. You need them, because Views are relative to a UserInterface and
           a Bookmark can, at run time, turn these into absolute URLs. Bookmarks also contain metadata,
           such as the title of the View it points to.
        """
        class UIWithRelativeView(UserInterface):
            def assemble(self):
                self.define_view('/aview', title='A View title')

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                ui_factory = self.define_user_interface('/a_ui',  UIWithRelativeView,  {}, name='myui')

                # How you could get a bookmark from a UserInterfaceFactory
                fixture.bookmark = ui_factory.get_bookmark(relative_path='/aview')

        wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
        Browser(wsgi_app).open('/a_ui/aview') # To execute the above once

        # What the bookmark knows
        vassert( fixture.bookmark.href.path == '/a_ui/aview' )
        vassert( fixture.bookmark.description == 'A View title' )
        vassert( fixture.bookmark.base_path == '/a_ui' )
        vassert( fixture.bookmark.relative_path == '/aview' )

        # How you would use a bookmark in other views (possibly in other UserInterfaces)
        a = A.from_bookmark(fixture.view, fixture.bookmark)
        
        # .. and how the A will be rendered
        a_etree = WidgetTester(a).render_html_tree()        
        vassert( a_etree.attrib['href'] == '/a_ui/aview' )
        vassert( a_etree.text == 'A View title' )

    class LifeCycleFixture(WebFixture):
        def current_view_is_plugged_in(self, page):
            return page.slot_contents['main_slot'].__class__ is Panel

    @test(LifeCycleFixture)
    def the_lifecycle_of_a_ui(self, fixture):
        """This test illustrates the steps a UserInterface goes through from being specified, to
           being used. It tests a couple of lower-level implementation issues (see comments)."""

        @stubclass(UserInterface)
        class UserInterfaceStub(UserInterface):
            assembled = False
            def assemble(self, **ui_arguments):
                self.controller_at_assemble_time = self.controller
                root = self.define_view('/some/path', title='A view')
                root.set_slot('slotA', Panel.factory())
                self.assembled = True

        # Phase1: specifying a user_interface and assembleing it to a site (with kwargs)
        parent_ui = None
#        parent_ui = EmptyStub(base_path='/')
        slot_map = {'slotA': 'main_slot'}
        ui_factory = UserInterfaceFactory(parent_ui, RegexPath('/', '/', {}), slot_map, UserInterfaceStub, 'test_ui')


        # Phase2: creating a user_interface
        request = Request.blank('/some/path')
        fixture.context.set_request(request)
        user_interface = ui_factory.create('/some/path', for_bookmark=False)

        # - Assembly happened correctly
        vassert( user_interface.parent_ui is parent_ui )
        vassert( user_interface.slot_map is slot_map )
        vassert( user_interface.name is 'test_ui' )
        vassert( user_interface.relative_base_path == '/' )
        vassert( user_interface.controller_at_assemble_time is not None)
        vassert( user_interface.controller is not None )
        vassert( user_interface.assembled )

        # - Create for_bookmark empties the relative_path
        user_interface = ui_factory.create('/some/path', for_bookmark=True)
        vassert( user_interface.relative_path == '' )

        # - The relative_path is reset if not done for_bookmark
        #   This is done in case a for_bookmark call just
        #   happened to be done for the same UserInterface in the same request
        #   before the "real" caal is done
        user_interface = ui_factory.create('/some/path', for_bookmark=False)
        vassert( user_interface.relative_path == '/some/path' )

        # Phase5: create the page and plug the view into it
        page = Widget.factory().create(user_interface.current_view)
        page.add_child(Slot(user_interface.current_view, 'main_slot'))

        page.plug_in(user_interface.current_view)
        vassert( fixture.current_view_is_plugged_in(page) )
        vassert( isinstance(user_interface.sub_resources, FactoryDict) )



# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import warnings
import itertools

import six
from six.moves import zip_longest
from nose.tools import istest
from reahl.tofu import expected
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert
from reahl.stubble import EmptyStub

from reahl.web.fw import UserInterface
from reahl.web.fw import Region
from reahl.web.ui import HTML5Page, P
from reahl.web.pure import PageColumnLayout
from reahl.webdev.tools import Browser
from reahl.web_dev.fixtures import WebFixture
from reahl.component.exceptions import ProgrammerError, IncorrectArgumentError, IsSubclass


class BasicScenarios(WebFixture):
    expected_warnings = []
    @scenario
    def view_with_page(self):
        class SimplePage(HTML5Page):
            def __init__(self, view):
                super(SimplePage, self).__init__(view)
                self.body.add_child(P(view, text='Hello world!'))

        class MainUI(UserInterface):
            def assemble(self):
                self.define_view('/', title='Hello', page=SimplePage.factory())

        self.MainUI = MainUI
        self.expected_content_length = 1602
        self.content_includes_p = True

    @scenario
    def view_with_set_page(self):
        class SimplePage(HTML5Page):
            def __init__(self, view):
                super(SimplePage, self).__init__(view)
                self.body.add_child(P(view, text='Hello world!'))

        class MainUI(UserInterface):
            def assemble(self):
                home = self.define_view('/', title='Hello')
                home.set_page(SimplePage.factory())

        self.MainUI = MainUI
        self.expected_content_length = 1602
        self.content_includes_p = True

    @scenario
    def ui_with_page(self):
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_view('/', title='Hello')

        self.MainUI = MainUI
        self.expected_content_length = 1583
        self.content_includes_p = False

    @scenario
    def backwards_compatibility(self):
        class MainUI(Region):
            def assemble(self):
                self.define_main_window(HTML5Page)
                self.define_view('/', title='Hello')

        self.MainUI = MainUI
        self.expected_content_length = 1583
        self.content_includes_p = False

        self.expected_warnings = ['Region has been renamed to UserInterface, please use UserInterface instead',
                                  'Region has been renamed to UserInterface, please use UserInterface instead', 
                                  'Please use .define_page() instead',
                                  'Region has been renamed to UserInterface, please use UserInterface instead',
                                  'Region has been renamed to UserInterface, please use UserInterface instead']


@test(BasicScenarios)
def basic_assembly(fixture):
    """An application is built by extending UserInterface, and defining this UserInterface in an .assemble() method.

    To define the UserInterface, several Views are defined. Views are mapped to URLs. When a user GETs
    the URL of a View, a page is rendered back to the user. How that page is created
    can happen in different ways, as illustrated by each scenario of this test.
    """
    wsgi_app = fixture.new_wsgi_app(site_root=fixture.MainUI)
    browser = Browser(wsgi_app)

    # GETting the URL results in the HTML for that View
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter('always')
        browser.open('/')
        vassert( browser.title == 'Hello' )

    warning_messages = [six.text_type(i.message) for i in caught_warnings]
    vassert( len(warning_messages) == len(fixture.expected_warnings) )
    for caught, expected_message in zip_longest(warning_messages, fixture.expected_warnings):
        vassert( expected_message in caught )

    if fixture.content_includes_p:
        [message] = browser.lxml_html.xpath('//p')
        vassert( message.text == 'Hello world!' )

    # The headers are set correctly
    response = browser.last_response
    vassert( response.content_length == fixture.expected_content_length )
    vassert( response.content_type == 'text/html' )
    vassert( response.charset == 'utf-8' )

    # Invalid URLs do not exist
    with warnings.catch_warnings(record=True):
        browser.open('/nonexistantview/', status=404)


@test(WebFixture)
def basic_error1(fixture):
    """Sending the the wrong kind of thing as widget_class to define_page is reported to the programmer."""
    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(EmptyStub)
            self.define_view('/', title='Hello')

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    with expected(IsSubclass):
        browser.open('/')

@test(WebFixture)
def basic_error2(fixture):
    """Sending the the wrong arguments for the specified class to define_page is reported to the programmer."""
    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page, 1, 2)
            self.define_view('/', title='Hello')

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    def check_exc(ex):
        msg = six.text_type(ex)
        vassert( msg.startswith('define_page was called with arguments that do not match those expected by') )
    with expected(IncorrectArgumentError, test=check_exc):
        browser.open('/')


@test(WebFixture)
def basic_error3(fixture):
    """Forgetting to define either a page of a page for a View is reported to the programmer."""
    class MainUI(UserInterface):
        def assemble(self):
            self.define_view('/', title='Hello')

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    def check_exc(ex):
        msg = six.text_type(ex)
        vassert( msg == 'there is no page defined for /' )
    with expected(ProgrammerError, test=check_exc):
        browser.open('/')


class SlotScenarios(WebFixture):
    @scenario
    def page_on_ui(self):
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
                home = self.define_view('/', title='Hello')
                home.set_slot('main', P.factory(text='Hello world'))
                home.set_slot('footer', P.factory(text='I am the footer'))
        self.MainUI = MainUI

    @scenario
    def page_on_view(self):
        class MainUI(UserInterface):
            def assemble(self):
                home = self.define_view('/', title='Hello')
                home.set_page(HTML5Page.factory().use_layout(PageColumnLayout('main')))
                home.set_slot('main', P.factory(text='Hello world'))
                home.set_slot('footer', P.factory(text='I am the footer'))
        self.MainUI = MainUI


@test(SlotScenarios)
def slots(fixture):
    """A View modifies the page by populating named Slots in the page with Widgets."""
    wsgi_app = fixture.new_wsgi_app(site_root=fixture.MainUI)
    browser = Browser(wsgi_app)
    
    browser.open('/')
    vassert( browser.title == 'Hello' )
    [main_p, footer_p] = browser.lxml_html.xpath('//p')
    vassert( main_p.text == 'Hello world' )
    vassert( footer_p.text == 'I am the footer' )


@test(WebFixture)
def slot_error(fixture):
    """Supplying contents for a slot that does not exist results in s sensible error."""
    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
            home = self.define_view('/', title='Hello')
            home.set_slot('main', P.factory(text='Hello world'))
            home.set_slot('nonexistantslotname', P.factory(text='I am breaking'))

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)

    def check_exc(ex):
        vassert( six.text_type(ex).startswith('An attempt was made to plug Widgets into the following slots that do not exist') )

    with expected(ProgrammerError, test=check_exc):
        browser.open('/')


@test(WebFixture)
def slot_defaults(fixture):
    """If a View does not specify contents for a Slot, the Slot will be populated by the window's default
       widget for that slot if specified, else it will be left empty.
    """
    class MainUI(UserInterface):
        def assemble(self):
            main = self.define_page(HTML5Page).use_layout(PageColumnLayout('main'))
            main.add_default_slot('main', P.factory(text='defaulted slot contents'))
            self.define_view('/', title='Hello')

    wsgi_app = fixture.new_wsgi_app(site_root=MainUI)
    browser = Browser(wsgi_app)
    
    browser.open('/')

    # The default widget for the main slot is used
    [p] = browser.lxml_html.xpath('//p')
    vassert( p.text == 'defaulted slot contents' )

    # The header slot has no default, and is thus left empty
    header_contents = browser.lxml_html.xpath('//header/*')
    vassert( not header_contents )




# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import expected

from reahl.web.fw import UserInterface
from reahl.web.ui import P, HTML5Page
from reahl.webdev.tools import Browser
from reahl.web_dev.fixtures import WebFixture
from reahl.component.exceptions import ProgrammerError

class UserInterfaceErrorScenarios(WebFixture):
    def new_wsgi_app(self):
        fixture = self
        class SimpleUserInterface(UserInterface):
            def assemble(self):
                root = self.define_view('/', title='View')
                root.set_slot('name', P.factory())

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_user_interface('/a_ui',  SimpleUserInterface,  {}, name='test_ui')

        return super(UserInterfaceErrorScenarios, self).new_wsgi_app(site_root=MainUI)

    @scenario
    def plug_in_to_nonexistant_name(self):
        self.slot_map = {'name': 'nonexistent'}

    @scenario
    def view_name_not_mapped(self):
        self.slot_map = {}

        
@istest
class UserInterfaceErrorTests(object):
    @test(UserInterfaceErrorScenarios)
    def ui_slots_map_error(self, fixture):
        browser = Browser(fixture.wsgi_app)

        with expected(ProgrammerError):
            browser.open('/a_ui/')





# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import re

from reahl.stubble import stubclass
from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert, expected

from reahl.webdev.tools import Browser
from reahl.web.fw import CheckedRemoteMethod
from reahl.web.fw import JsonResult
from reahl.web.fw import MethodResult
from reahl.web.fw import RemoteMethod
from reahl.web.fw import Widget
from reahl.web.fw import WidgetResult
from reahl.component.modelinterface import Field, IntegerField
from reahl.component.py3compat import ascii_as_bytes_or_str
from reahl.web_dev.fixtures import  WebFixture

class RemoteMethodFixture(WebFixture):
    def new_widget_factory(self, remote_method=None):
        remote_method = remote_method or self.remote_method
        
        @stubclass(Widget)
        class WidgetWithRemoteMethod(Widget):
            def __init__(self, view):
                super(WidgetWithRemoteMethod, self).__init__(view)
                view.add_resource(remote_method)
                
        return WidgetWithRemoteMethod.factory()

    def new_wsgi_app(self, remote_method=None):
        remote_method = remote_method or self.remote_method
        slot_definitions = {'main': self.new_widget_factory(remote_method=remote_method)}
        return super(RemoteMethodFixture, self).new_wsgi_app(view_slots=slot_definitions)
            
        
@istest
class RemoteMethodTests(object):
    @test(WebFixture)
    def remote_methods(self, fixture):
        """A RemoteMethod is a SubResource representing a method on the server side which can be invoked via POSTing to an URL."""

        def callable_object():
            return 'value returned from method'

        encoding = 'koi8_r'  # Deliberate
        remote_method = RemoteMethod('amethod', callable_object, MethodResult(mime_type='ttext/hhtml', encoding=encoding))
    
        @stubclass(Widget)
        class WidgetWithRemoteMethod(Widget):
            def __init__(self, view):
                super(WidgetWithRemoteMethod, self).__init__(view)
                view.add_resource(remote_method)

        wsgi_app = fixture.new_wsgi_app(view_slots={'main': WidgetWithRemoteMethod.factory()})
        browser = Browser(wsgi_app)
        
        # By default you cannot GET, since the method is not immutable
        browser.open('/_amethod_method', status=405)
        
        # POSTing to the URL, returns the result of the method
        browser.post('/_amethod_method', {})
        vassert( browser.raw_html == 'value returned from method' )
        vassert( browser.last_response.charset == encoding)
        vassert( browser.last_response.content_type == 'ttext/hhtml' )
    
    @test(RemoteMethodFixture)
    def exception_handling(self, fixture):
        """The RemoteMethod sends back the six.text_type() of an exception raised for the specified exception class."""
        
        def fail():
            raise Exception('I failed')
        remote_method = RemoteMethod('amethod', fail, MethodResult(catch_exception=Exception))

        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)

        browser.post('/_amethod_method', {})
        vassert( browser.raw_html == 'I failed' )

    @test(RemoteMethodFixture)
    def immutable_remote_methods(self, fixture):
        """A RemoteMethod that is immutable is accessible via GET (instead of POST)."""

        def callable_object():
            return 'value returned from method'
        remote_method = RemoteMethod('amethod', callable_object, MethodResult(), immutable=True)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        # GET, since the method is immutable
        browser.open('/_amethod_method')
        vassert( browser.raw_html == 'value returned from method' )
        
        # POSTing to the URL, is not supported
        browser.post('/_amethod_method', {}, status=405)
    
    
    class ArgumentScenarios(RemoteMethodFixture):
        @scenario
        def get(self):
            self.immutable = True
        @scenario
        def post(self):
            self.immutable = False
            
    @test(ArgumentScenarios)
    def arguments_to_remote_methods(self, fixture):
        """A RemoteMethod can get arguments from a query string or submitted form values, depending on the scenario."""

        def callable_object(**kwargs):
            fixture.method_kwargs = kwargs 
            return ''
        remote_method = RemoteMethod('amethod', callable_object, MethodResult(), immutable=fixture.immutable)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        kwargs_sent = {'a':'AAA', 'b':'BBB'}
        if fixture.immutable:
            browser.open('/_amethod_method?a=AAA&b=BBB')
        else:
            browser.post('/_amethod_method', kwargs_sent)
        vassert( fixture.method_kwargs == kwargs_sent )

    @test(ArgumentScenarios)
    def checked_arguments(self, fixture):
        """A CheckedRemoteMethod checks and marshalls its parameters using Fields."""

        def callable_object(anint=None, astring=None):
            fixture.method_kwargs = {'anint': anint, 'astring': astring}
            return ''
        remote_method = CheckedRemoteMethod('amethod', callable_object, MethodResult(), 
                                            immutable=fixture.immutable, 
                                            anint=IntegerField(),
                                            astring=Field())
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        if fixture.immutable:
            browser.open('/_amethod_method?anint=5&astring=SupercalifraGilisticexpialidocious')
        else:
            browser.post('/_amethod_method', {'anint':'5', 'astring':'SupercalifraGilisticexpialidocious'})
        vassert( fixture.method_kwargs == {'anint':5, 'astring':'SupercalifraGilisticexpialidocious'} )

    class ResultScenarios(RemoteMethodFixture):
        @scenario
        def json(self):
            self.method_result = JsonResult(IntegerField(), catch_exception=Exception)
            self.value_to_return = 1
            self.expected_response = '1'
            self.exception_response = '"exception text"'
            self.expected_charset = self.method_result.encoding
            self.expected_content_type = 'application/json'

        @scenario
        def widget(self):
            @stubclass(Widget)
            class WidgetStub(object):
                css_id = 'someid'
                def render_contents(self): return '<the widget contents>'
                def get_contents_js(self, context=None): return ['some', 'some', 'javascript']

            self.method_result = WidgetResult(WidgetStub())
            self.value_to_return = 'ignored in this case'
            self.expected_response = '<the widget contents><script type="text/javascript">javascriptsome</script>'
            self.exception_response = Exception
            self.expected_charset = self.method_result.encoding
            self.expected_content_type = 'text/html'
            
    @test(ResultScenarios)
    def different_kinds_of_result(self, fixture):
        """Different kinds of MethodResult can be specified for a method."""

        def callable_object():
            return fixture.value_to_return
        remote_method = RemoteMethod('amethod', callable_object, default_result=fixture.method_result)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        browser.post('/_amethod_method', {})
        vassert( re.match(fixture.expected_response, browser.raw_html) )
        vassert( browser.last_response.charset == fixture.expected_charset )
        vassert( browser.last_response.content_type == fixture.expected_content_type )

    @test(ResultScenarios.json)
    def exception_handling_for_json(self, fixture):
        """How exceptions are handled with JsonResult."""

        def fail():
            raise Exception('exception text')
        remote_method = RemoteMethod('amethod', fail, default_result=fixture.method_result)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)
        
        browser.post('/_amethod_method', {})
        vassert( browser.raw_html == fixture.exception_response )
        vassert( browser.last_response.charset == fixture.expected_charset )
        vassert( browser.last_response.content_type == fixture.expected_content_type )

    @test(ResultScenarios.widget)
    def exception_handling_for_widgets(self, fixture):
        """How exceptions are handled with WidgetResult."""

        def fail():
            raise Exception('exception text')
        remote_method = RemoteMethod('amethod', fail, default_result=fixture.method_result)
    
        wsgi_app = fixture.new_wsgi_app(remote_method=remote_method)
        browser = Browser(wsgi_app)

        with expected(Exception):
            browser.post('/_amethod_method', {})




# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division
import six
import os
from six.moves import http_cookies 

from webob import Request, Response

from reahl.stubble import stubclass
from reahl.tofu import Fixture
from reahl.sqlalchemysupport import Session

from reahl.web.fw import ComposedPage, ReahlWSGIApplication, WebExecutionContext, \
                         UserInterfaceFactory, IdentityDictionary, FactoryDict, UrlBoundView, UserInterface, \
                         WidgetList, Url, Widget, RegexPath
from reahl.web.ui import HTML5Page
from reahl.web.pure import PageColumnLayout
from reahl.component.i18n import Translator
from reahl.component.py3compat import ascii_as_bytes_or_str
from reahl.domain_dev.fixtures import PartyModelZooMixin
from reahl.domain.systemaccountmodel import LoginSession
from reahl.webdev.tools import DriverBrowser
from reahl.webdeclarative.webdeclarative import UserSession


_ = Translator('reahl-webdev')

        
@stubclass(ReahlWSGIApplication)
class ReahlWSGIApplicationStub(ReahlWSGIApplication):
    def add_reahl_static_files(self): # To save time, this is costly...
        pass  

    
class WebBasicsMixin(PartyModelZooMixin):
    def log_in(self, browser=None, session=None, system_account=None, stay_logged_in=False):
        session = session or self.session
        browser = browser or self.driver_browser
        login_session = LoginSession.for_session(session)
        login_session.set_as_logged_in(system_account or self.system_account, stay_logged_in)
        # quickly create a response so the fw sets the cookies, which we copy and explicitly set on selenium.
        response = Response()
        self.session.set_session_key(response)
        cookies = http_cookies.BaseCookie(ascii_as_bytes_or_str(', '.join(response.headers.getall('set-cookie'))))
        for name, morsel in cookies.items():
            cookie = {'name':name, 'value':morsel.value}
            cookie.update(dict([(key, value) for key, value in morsel.items() if value]))
            browser.create_cookie(cookie)
    
    def new_driver_browser(self, driver=None):
        driver = driver or self.web_driver
        return DriverBrowser(driver)

    @property
    def chrome_driver(self):
        return self.run_fixture.chrome_driver
    @property
    def firefox_driver(self):
        return self.run_fixture.firefox_driver
    @property
    def web_driver(self):
        return self.run_fixture.web_driver
    @property
    def reahl_server(self):
        return self.run_fixture.reahl_server
        
    def new_context(self, request=None, config=None, session=None):
        context = WebExecutionContext()
        context.set_config( config or self.config )
        context.set_system_control( self.system_control )
        context.set_request( request or self.request )
        with context:
            context.set_session( session or self.session )
        return context
        
    def new_session(self):
        web_user_session = UserSession()
        Session.add(web_user_session)
        return web_user_session

    def new_request(self, path=None, url_scheme=None):
        request = Request.blank(path or '/', charset='utf8')
        request.environ['wsgi.url_scheme'] = url_scheme or 'http'
        if request.scheme == 'http':
            request.environ['SERVER_PORT'] = '8000'
            request.host = 'localhost:8000'
        else:
            request.environ['SERVER_PORT'] = '8363'
            request.host = 'localhost:8363'
        return Request(request.environ, charset='utf8')

    def new_wsgi_app(self, site_root=None, enable_js=False, 
                         config=None, view_slots=None, child_factory=None):
        wsgi_app_class = ReahlWSGIApplicationStub
        if enable_js:
            wsgi_app_class = ReahlWSGIApplication
        view_slots = view_slots or {}
        child_factory = child_factory or Widget.factory()
        if 'main' not in view_slots:
            view_slots['main'] = child_factory
        config = config or self.config

        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(PageColumnLayout(*view_slots.keys()))
                self.define_view('/', title='Home page', slot_definitions=view_slots)

        site_root = site_root or MainUI
        config.web.site_root = site_root
        return wsgi_app_class(config)

    @property
    def current_location(self):
        return Url(self.driver_browser.get_location())

    def new_view(self):
        current_path = Url(WebExecutionContext.get_context().request.url).path
        view = UrlBoundView(self.user_interface, current_path, 'A view', {})
        return view

    def new_user_interface(self):
        return UserInterface(None, '/', {}, False, 'test_ui')


class WebFixture(Fixture, WebBasicsMixin):
    pass






# Copyright 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

from reahl.tofu import test
from reahl.tofu import vassert

from reahl.webdev.tools import XPath
from reahl.web.ui import Panel, Button, StaticColumn, DynamicColumn
from reahl.web_dev.widgets.tabletests import TableFixture

from reahl.web.table import DataTable
from reahl.component.modelinterface import Field, IntegerField, exposed



class DataItem(object):
    row = None
    alpha = None

    def __init__(self, row, alpha):
        self.row = row
        self.alpha = alpha
    
    @exposed
    def fields(self, fields):
        fields.row = IntegerField(label='Row', required=True, default = self.row)
        fields.alpha = Field(label='Alpha', required=True, default = self.alpha)

        
class DataTableFixture(TableFixture):

    items_per_page = 3
        
    def new_columns(self):
        return [StaticColumn(IntegerField(label='Row Number'), 'row', sort_key=lambda i: i.row),
                StaticColumn(Field(label='Alpha'), 'alpha', sort_key=lambda i: i.alpha),
                ]

    def new_data(self):
        alphas = 'THEQUICKBROWNFXJMPSVLAZYDG'
        return [DataItem(number, alpha) for number, alpha in enumerate(alphas, start=1)]
        #[(1, 'T'), (2, 'H'), (3, 'E'), (4, 'Q'),...(23, 'Z'), (24, 'Y'), (25, 'D'), (26, 'G')]

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Panel):
            def __init__(self, view):
                super(MainWidget, self).__init__(view)

                data_table = DataTable(view,
                                 fixture.columns,
                                 fixture.data,
                                 items_per_page=fixture.items_per_page,
                                 caption_text='All my friends',
                                 summary='Summary for screen reader',
                                 css_id='my_table_data')
                self.add_child(data_table)
        return MainWidget

    def new_wsgi_app(self):
        return super(TableFixture, self).new_wsgi_app(enable_js=True,
                                                      child_factory=self.MainWidget.factory())

    def xpath_for_ascending_link_for_column(self, column_number):
        return '(((//table/thead/tr/th)[%s]/span)[2]/a)[1]' % column_number

    def xpath_for_descending_link_for_column(self, column_number):
        return '(((//table/thead/tr/th)[%s]/span)[2]/a)[2]' % column_number

    def get_table_header(self):
        header_data = []
        for column_number in list(range(1, len(self.columns)+1)):
            header_data.append(self.driver_browser.web_driver.find_element_by_xpath('(//table/thead/tr/th)[%s]' % (column_number)).text)
        return header_data


@test(DataTableFixture)
def paging_through_data(fixture):
    """DataTable splits its items into different pages (between which a user can navigate), showing only the items of a particular page at a time."""
    fixture.reahl_server.set_app(fixture.wsgi_app)
    fixture.driver_browser.open('/')

    #click to last page
    fixture.driver_browser.click(XPath.link_with_text('>|'))
    fixture.driver_browser.click(XPath.link_with_text('9'))
    
    vassert( fixture.table_number_rows() == 2 )
    vassert( fixture.get_table_row(1) == ['25' ,'D'] )
    vassert( fixture.get_table_row(2) == ['26' ,'G'] )

    #click to page 4
    fixture.driver_browser.click(XPath.link_with_text('|<'))
    fixture.driver_browser.click(XPath.link_with_text('4'))
    
    vassert( fixture.table_number_rows() == 3 )
    vassert( fixture.get_table_row(1) == ['10' ,'R'] )
    vassert( fixture.get_table_row(2) == ['11' ,'O'] )
    vassert( fixture.get_table_row(3) == ['12' ,'W'] )


@test(DataTableFixture)
def sorting(fixture):
    """By clicking on special links in the column header, the table is sorted according to that column - ascending or descending."""
    fixture.reahl_server.set_app(fixture.wsgi_app)
    fixture.driver_browser.open('/')

    #----- by default, not sorted
    vassert( fixture.get_table_row(1) == ['1' ,'T'] )
    vassert( fixture.get_table_row(2) == ['2' ,'H'] )
    vassert( fixture.get_table_row(3) == ['3' ,'E'] )
    
    #----- sort ascending on alpha, the second column
    fixture.driver_browser.click(fixture.xpath_for_ascending_link_for_column(2))
    
    vassert( fixture.get_table_row(1) == ['22' ,'A'] )
    vassert( fixture.get_table_row(2) == ['9' ,'B'] )
    vassert( fixture.get_table_row(3) == ['7' ,'C'] )

    #----- sort descending on alpha, the second column
    fixture.driver_browser.click(fixture.xpath_for_descending_link_for_column(2))
    
    vassert( fixture.get_table_row(1) == ['23' ,'Z'] )
    vassert( fixture.get_table_row(2) == ['24' ,'Y'] )
    vassert( fixture.get_table_row(3) == ['15' ,'X'] )

    #----- sort order stays changed when paging
    fixture.driver_browser.click(XPath.link_with_text('4'))
    
    vassert( fixture.get_table_row(1) == ['4' ,'Q'] )
    vassert( fixture.get_table_row(2) == ['18' ,'P'] )
    vassert( fixture.get_table_row(3) == ['11' ,'O'] )

    #----- contents of the page you are on changes according to a new sort order
    fixture.driver_browser.click(fixture.xpath_for_ascending_link_for_column(1))
    
    vassert( fixture.get_table_row(1) == ['10' ,'R'] )
    vassert( fixture.get_table_row(2) == ['11' ,'O'] )
    vassert( fixture.get_table_row(3) == ['12' ,'W'] )


@test(DataTableFixture)
def which_columns_can_cause_sorting(fixture):
    """Only columns with sort_key specified are sortable"""

    fixture.columns.append(StaticColumn(Field(label='Not sortable'), 'alpha'))

    fixture.reahl_server.set_app(fixture.wsgi_app)
    fixture.driver_browser.open('/')

    vassert( fixture.get_table_header() == ['Row Number▲▼', 'Alpha▲▼', 'Not sortable'] )





# Copyright 2015 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

import pkg_resources

from reahl.tofu import test
from reahl.tofu import temp_dir
from reahl.tofu import temp_file_with
from reahl.tofu import vassert
from reahl.stubble import easter_egg, stubclass

from reahl.web.fw import ReahlWSGIApplication, UserInterface
from reahl.web.ui import HTML5Page
from reahl.web.libraries import Library
from reahl.web.egg import WebConfig
from reahl.component.config import Configuration, ReahlSystemConfig
from reahl.sqlalchemysupport import SqlAlchemyControl
from reahl.webdev.tools import Browser
from reahl.web_dev.fixtures import WebFixture


class LibraryFixture(WebFixture):
    def new_egg_dir(self):
        egg_dir = temp_dir()
        package_dir = egg_dir.sub_dir('static_files')
        init_file = package_dir.file_with('__init__.py', '')
        js_file = package_dir.file_with('somefile.js', 'contents - js')
        css_file = package_dir.file_with('somefile.css', 'contents - css')
        return egg_dir

    def new_MyLibrary(self):
        easter_egg.clear()
        pkg_resources.working_set.add(easter_egg)
        easter_egg.set_module_path(self.egg_dir.name)

        class MyLibrary(Library):
            def __init__(self):
                super(MyLibrary, self).__init__('mylib')
                self.files = ['somefile.js', 'somefile.css']
                self.shipped_in_directory = '/static_files'
                self.egg_name = easter_egg.project_name

        return MyLibrary

    def new_config(self):
        config = Configuration()
        config.reahlsystem = self.new_reahlsystem()
        config.web = self.new_webconfig()
        config.web.site_root = self.MainUI
        return config

    def new_MainUI(self):
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page)
                self.define_view('/', title='Home page')
        return MainUI



@test(LibraryFixture)
def configuring_libraries(fixture):
    """Reahl can be configured to expose frontend libraries (libraries of js and css files)."""
    web = fixture.config.web
    vassert( 'mylib' not in web.frontend_libraries )
    web.frontend_libraries.add(fixture.MyLibrary())
    vassert( 'mylib' in web.frontend_libraries )


@test(LibraryFixture)
def library_files(fixture):
    """The files part of configured frontend libraries are (a) added to /static and also (b) included on any page."""
    fixture.config.web.frontend_libraries.clear()
    fixture.config.web.frontend_libraries.add(fixture.MyLibrary())

    browser = Browser(ReahlWSGIApplication(fixture.config))

    browser.open('/static/somefile.js')
    vassert( browser.raw_html == 'contents - js' )

    browser.open('/static/somefile.css')
    vassert( browser.raw_html == 'contents - css' )

    browser.open('/')
    script_added = browser.get_html_for('//script')
    vassert( script_added == '<script type="text/javascript" src="/static/somefile.js"></script>\n' )

    link_added = browser.get_html_for('//link')
    vassert( link_added == '<link rel="stylesheet" href="/static/somefile.css" type="text/css">\n' )

@test(WebFixture)
def standard_reahl_files(fixture):
    """The framework includes certain frontent frameworks by default."""
    wsgi_app = ReahlWSGIApplication(fixture.config)
    browser = Browser(wsgi_app)

    browser.open('/static/html5shiv-printshiv-3.6.3.js')
    vassert( browser.last_response.content_length > 0 )

    browser.open('/static/IE9.js')
    vassert( browser.last_response.content_length > 0 )

    browser.open('/static/reahl.js')
    vassert( browser.last_response.content_length > 0 )

    browser.open('/static/reahl.css')
    vassert( browser.last_response.content_length > 0 )

    browser.open('/static/runningon.png')
    vassert( browser.last_response.content_length > 0 )




# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import six
from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert

from reahl.web.fw import Url
from reahl.web.ui import Form, TextInput
from reahl.component.modelinterface import EmailField, exposed

from reahl.web_dev.inputandvalidation.eventhandling import FormFixture
from reahl.webdev.tools import Browser


class Scenarios(FormFixture):
    @scenario
    def valid_field(self):
        # - a field that passes validation
        self.url = Url('/__some_form_validate_method?field_name=valid@email.org')
        self.expected_body = 'true'
        self.expected_status = '200 OK' 
        self.expected_content_type = 'application/json'
        self.expected_charset = 'utf-8'

    @scenario
    def failing_field(self):
        # - a field that fails one or more? constraints
        self.url = Url('/__some_form_validate_method?field_name=invalidaddress')
        self.expected_body = '"field_name should be a valid email address"'
        self.expected_status = '200 OK' 
        self.expected_content_type = 'application/json'
        self.expected_charset = 'utf-8'

    @scenario
    def non_existent_field(self):
        # - a field that does not exist
        self.url = Url('/__some_form_validate_method?nonexistantfield=value')
        self.expected_body = 'false'
        self.expected_status = '200 OK' 
        self.expected_content_type = 'application/json'
        self.expected_charset = 'utf-8'

    @scenario
    def empty_querystring(self):
        # - an empty querystring
        self.url = Url('/__some_form_validate_method')
        self.expected_body = 'false'
        self.expected_status = '200 OK' 
        self.expected_content_type = 'application/json'
        self.expected_charset = 'utf-8'

    
@istest
class FieldValidatorTests(object):
    @test(Scenarios)
    def remote_field_validator_handles_GET(self, fixture):
        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.field_name = EmailField()

        model_object = ModelObject()

        class MyForm(Form):
            def __init__(self, view, name):
                super(MyForm, self).__init__(view, name)
                self.add_child(TextInput(self, model_object.fields.field_name))

        wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory(name='some_form'))
        fixture.reahl_server.set_app(wsgi_app)
        browser = Browser(wsgi_app)

        browser.open(six.text_type(fixture.url))
        response = browser.last_response

        vassert( response.unicode_body == fixture.expected_body )
        vassert( response.status == fixture.expected_status )
        vassert( response.content_type == fixture.expected_content_type )
        vassert( response.charset == fixture.expected_charset )




# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert, expected
from reahl.stubble import EmptyStub

from reahl.web.ui import Input, TextInput, Button, Form, ValidationException, \
                          LabelOverInput, CueInput, CheckboxInput, TextInput, InputLabel, ButtonInput,\
                          PasswordInput, Button, LabelledInlineInput, LabelledBlockInput, P,\
                          TextArea, SelectInput, RadioButtonInput

from reahl.component.modelinterface import Field, EmailField, BooleanField,\
                             RequiredConstraint, MinLengthConstraint, PatternConstraint, RemoteConstraint,\
                             Event, Allowed, exposed, Action, Choice, ChoiceGroup, ChoiceField, IntegerField,\
                             MultiChoiceField, DateField
from reahl.component.exceptions import IsInstance
from reahl.web_dev.fixtures import WebFixture
from reahl.webdev.tools import WidgetTester
from reahl.webdev.tools import XPath
from reahl.component_dev.fieldtests import FieldMixin


class InputMixin(FieldMixin):
    def new_form(self):
        return Form(self.view, 'test')

    def new_event(self, label='click me'):
        event = Event(label=label)
        event.bind('aname', None)
        self.form.define_event_handler(event)
        return event


class SimpleInputFixture(WebFixture, InputMixin):
    pass


class InputMixin2(InputMixin):
    def new_field(self, label='the label'):
        return Field(label=label)
    def new_model_object(self):
        fixture = self
        class ModelObject(object):
            def handle_event(self):
                pass
            @exposed
            def events(self, events):
                events.an_event = Event(label='click me', action=Action(self.handle_event))
            @exposed
            def fields(self, fields):
                fields.an_attribute = fixture.field
        return ModelObject()


class InputStateFixture(WebFixture, InputMixin):
    def new_field(self):
        field = EmailField(default='default value')
        field.bind('an_attribute', self.model_object)
        return field

    def new_input(self):    
        return Input(self.form, self.field)

    @scenario
    def defaulted_input_no_value_on_domain(self):
        del self.model_object.an_attribute
        self.input.prepare_input()

        self.expected_state = 'defaulted'
        self.expected_value = 'default value'

    @scenario
    def defaulted_input_value_on_domain(self):
        self.input.prepare_input()
        self.model_object.an_attribute = 'something'

        self.expected_state = 'defaulted'
        self.expected_value = 'something'

    @scenario
    def validly_entered(self):
        self.input.enter_value('someone@home.org')
        self.input.prepare_input()

        self.expected_state = 'validly_entered' 
        self.expected_value = 'someone@home.org'

    @scenario
    def empty_input_non_required_field(self):
        self.input.enter_value('')
        self.input.prepare_input()

        self.expected_state = 'validly_entered' 
        self.expected_value = ''

    @scenario
    def empty_input_required_field(self):
        self.field.make_required('')
        self.input.enter_value('')
        self.input.prepare_input()

        self.expected_state = 'invalidly_entered' 
        self.expected_value = ''

    @scenario
    def invalidly_entered(self):
        self.input.enter_value('not an email address')
        self.input.prepare_input()

        self.expected_state = 'invalidly_entered' 
        self.expected_value = 'not an email address'


@test(InputStateFixture)
def the_states_of_an_input(fixture):
    """An Input can be in one of three states: defaulted, invalidly_entered or validly_entered. Depending
       on the state of the input, it will have a different value when rendering.

       The states:
         - defaulted: the user has not entered anything
           value:     the value on the model object which is bound to its Field if that object
                      has such an an_attribute, else the default value of the Field.
         - invalidly_entered: the user entered a value which failed validation
           value:             the invalid value which the user entered
         - validly_entered: the user entered a valid value
           value:           the value which the user entered even if the underlying
                            model object has a corresponding value
    """
    vassert( fixture.input.get_input_status() == fixture.expected_state )
    vassert( fixture.input.value == fixture.expected_value )


class Scenarios(WebFixture, InputMixin):
    @scenario
    def input_label(self):
        html_input = TextInput(self.form, self.field)
        self.widget = InputLabel(html_input)
        self.expected_html = '<label for="an_attribute">the label</label>'
        self.field_controls_visibility = True

    @scenario
    def input_label_with_text(self):
        html_input = TextInput(self.form, self.field)
        self.widget = InputLabel(html_input, text='some text')
        self.expected_html = '<label for="an_attribute">some text</label>'
        self.field_controls_visibility = True

    @scenario
    def button_input(self):
        self.widget = self.form.add_child(ButtonInput(self.form, self.event))
        self.expected_html = '<input name="event.aname?" form="test" type="submit" value="click me">'
        self.field_controls_visibility = False

    @scenario
    def button(self):
        self.widget = self.form.add_child(Button(self.form, self.event))
        self.expected_html = '<span class="reahl-button"><input name="event.aname?" form="test" type="submit" value="click me"></span>'
        self.field_controls_visibility = False

    @scenario
    def labelled_inline_input(self):
        self.model_object.an_attribute = 'field value'
        self.widget = self.form.add_child(LabelledInlineInput(TextInput(self.form, self.field)))
        self.expected_html = '<span class="reahl-labelledinput">'\
                             '<label for="an_attribute">the label</label>' \
                             '<span><input name="an_attribute" form="test" type="text" value="field value" class="reahl-textinput"></span>' \
                             '</span>'
        self.field_controls_visibility = True

    @scenario
    def labelled_block_input(self):
        self.model_object.an_attribute = 'field value'
        self.widget = self.form.add_child(LabelledBlockInput(TextInput(self.form, self.field)))
        self.expected_html = '<div class="pure-g reahl-labelledinput">'\
                             '<div class="column-label pure-u-1-4"><label for="an_attribute">the label</label></div>' \
                             '<div class="column-input pure-u-3-4"><input name="an_attribute" form="test" type="text" value="field value" class="reahl-textinput"></div>' \
                             '</div>'
        self.field_controls_visibility = True

    @scenario
    def password(self):
        self.widget = self.form.add_child(PasswordInput(self.form, self.field))
        self.expected_html = '<input name="an_attribute" form="test" type="password">'
        self.field_controls_visibility = True

    @scenario
    def checkbox_true(self):
        """A checkbox needs a 'checked' an_attribute if its field is True. It also renders ONLY the
            validation message for its required validation_constraint, not for all constraints"""
        self.model_object.an_attribute = True

        self.field = BooleanField(required=True, label='my text', required_message='$label is needed here')
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(CheckboxInput(self.form, self.field))
        self.expected_html = '<input name="an_attribute" checked="checked" form="test" required="*" type="checkbox" ' \
                             'class="{&quot;validate&quot;: {&quot;messages&quot;: {&quot;required&quot;: &quot;my text is needed here&quot;}}}">'
        self.field_controls_visibility = True

    @scenario
    def checkbox_false(self):
        self.checkbox_true()
        self.model_object.an_attribute = False
        self.expected_html = '<input name="an_attribute" form="test" required="*" type="checkbox" '\
                             'class="{&quot;validate&quot;: {&quot;messages&quot;: {&quot;required&quot;: &quot;my text is needed here&quot;}}}">'
        self.field_controls_visibility = True

    @scenario
    def text_area_input(self):
        self.widget = self.form.add_child(TextArea(self.form, self.field, rows=30, columns=20))
        self.expected_html = '<textarea name="an_attribute" cols="20" rows="30">field value</textarea>'
        self.field_controls_visibility = True

    @scenario
    def select_input(self):
        self.model_object.an_attribute = 2

        choices = [Choice(False, BooleanField(label='None')),
                   ChoiceGroup('grouped', [
                       Choice(1, IntegerField(label='One')),
                       Choice(2, IntegerField(label='Two'))])
                  ]
        self.field = ChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(SelectInput(self.form, self.field))
        group = '<optgroup label="grouped"><option value="1">One</option><option selected="selected" value="2">Two</option></optgroup>'
        option = '<option value="off">None</option>'
        self.expected_html = '<select name="an_attribute" form="test">%s%s</select>' % (option, group)
        self.field_controls_visibility = True

    @scenario
    def select_input_multi(self):
        self.model_object.an_attribute = [2]

        choices = [Choice(1, IntegerField(label='One')),
                   Choice(2, IntegerField(label='Two'))
                  ]
        self.field = MultiChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(SelectInput(self.form, self.field))
        options = '<option value="1">One</option><option selected="selected" value="2">Two</option>'
        self.expected_html = '<select name="an_attribute" form="test" multiple="multiple">%s</select>' % (options)
        self.field_controls_visibility = True

    @scenario
    def radio_button(self):
        self.model_object.an_attribute = 2

        choices = [ Choice(1, IntegerField(label='One')),
                    Choice(2, IntegerField(label='Two')) ]
        self.field = ChoiceField(choices)
        self.field.bind('an_attribute', self.model_object)

        self.widget = self.form.add_child(RadioButtonInput(self.form, self.field))

        radio_button = '<span class="reahl-radio-button"><input name="an_attribute"%s form="test" type="radio" value="%s">%s</span>'

        outer_div = '<div class="reahl-radio-button-input">%s</div>'
        buttons = (radio_button % ('', '1', 'One')) + (radio_button % (' checked="checked"', '2', 'Two'))
        self.expected_html = outer_div % buttons
        self.field_controls_visibility = True


@test(Scenarios)
def basic_rendering(fixture):
    """What the rendered html for a number of simple inputs look like."""

    tester = WidgetTester(fixture.widget)

    # Case: Normal behaviour
    actual = tester.render_html()
    vassert( actual == fixture.expected_html )

    # Case: When the Widget should not be visible
    fixture.field.access_rights.readable = Allowed(False)
    fixture.field.access_rights.writable = Allowed(False)
    actual = tester.render_html()
    if fixture.field_controls_visibility:
        vassert( actual == '' )
    else:
        vassert( actual == fixture.expected_html )


@test(SimpleInputFixture)
def wrong_args_to_input(fixture):
    """Passing the wrong arguments upon constructing an Input results in an error."""

    with expected(IsInstance):
        Input(fixture.form, EmptyStub())

    with expected(IsInstance):
        Input(EmptyStub(), Field())


class CheckboxFixture(WebFixture, InputMixin2):
    def new_field(self):
        return BooleanField(label='my text')


@test(CheckboxFixture)
def marshalling_of_checkbox(fixture):
    """When a form is submitted, the value of a checkbox is derived from
       whether the checkbox is included in the submission or not."""

    model_object = fixture.model_object
    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            checkbox = self.add_child(CheckboxInput(self, model_object.fields.an_attribute))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))
            fixture.checkbox = checkbox

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'))
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # Case: checkbox is submitted with form (ie checked)
    fixture.driver_browser.check("//input[@type='checkbox']")
    fixture.driver_browser.click("//input[@value='click me']")

    vassert( model_object.an_attribute )
    vassert( fixture.checkbox.value == 'on' )

    # Case: checkbox is not submitted with form (ie unchecked)
    fixture.driver_browser.uncheck("//input[@type='checkbox']")
    fixture.driver_browser.click("//input[@value='click me']")

    vassert( not model_object.an_attribute )
    vassert( fixture.checkbox.value == 'off' )


class LabelOverInputFixture(WebFixture, InputMixin2):
    label_xpath = "//form/span[@class='reahl-labelledinput reahl-labeloverinput']/label[@for='an_attribute' and text()='the label']"


@test(LabelOverInputFixture)
def label_behaviour(fixture):
    """The label is displayed over the input, but it disappears when selected
       or when there's a value in the input."""

    model_object = fixture.model_object

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(LabelOverInput(TextInput(self, model_object.fields.an_attribute)))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    fixture.reahl_server.set_app(wsgi_app)

    # Case: the field is empty and does not have a default value
    # - render initially with label
    fixture.driver_browser.open('/')
    vassert( fixture.driver_browser.is_element_present(fixture.label_xpath) )
    fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)

    # - if click, label disappears, input gets focus
    fixture.driver_browser.click(fixture.label_xpath)
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)

    # - if focus lost, label appears again if nothing is entered
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)

    # - if focus lost, label does not appear again if something is entered
    fixture.driver_browser.type('//input[@type="text"]', 'my@email.com')
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)


    # Case: the field has a defaulted value
    model_object.an_attribute = 'some value'
    fixture.driver_browser.open('/')

    # - render initially without a label
    fixture.driver_browser.wait_for_element_not_visible(fixture.label_xpath)

    # - if value removed, and focus lost, label appears
    fixture.driver_browser.type('//input[@type="text"]', '')
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_visible(fixture.label_xpath)


@test(LabelOverInputFixture)
def basic_rendering_label_over(fixture):
    """What the html for a LabelOverInput looks like."""

    fixture.field.bind('an_attribute', fixture.model_object)
    label_over_input = fixture.form.add_child(LabelOverInput(TextInput(fixture.form, fixture.field)))
    tester = WidgetTester(label_over_input)

    # Case: HTML - with a value
    #  - has CSS class reahl-labeloverinput
    #  - the label is hidden
    fixture.model_object.an_attribute = 'my value'
    actual = tester.render_html()
    expected_html = '<span class="reahl-labelledinput reahl-labeloverinput">'\
               '<label for="an_attribute" hidden="true">the label</label>' \
               '<span><input name="an_attribute" form="test" type="text" value="my value" class="reahl-textinput"></span>'\
               '</span>'
    vassert( actual == expected_html )

    # Case: HTML - without a value
    #  - has CSS class reahl-labeloverinput
    #  - the label is not hidden
    del fixture.model_object.an_attribute
    actual = tester.render_html()
    expected_html = '<span class="reahl-labelledinput reahl-labeloverinput">'\
               '<label for="an_attribute">the label</label>' \
               '<span><input name="an_attribute" form="test" type="text" value="" class="reahl-textinput"></span>'\
               '</span>'
    vassert( actual == expected_html )

    # Case: when the underlying input is not visible
    fixture.field.access_rights.readable = Allowed(False)
    fixture.field.access_rights.writable = Allowed(False)
    actual = tester.render_html()
    vassert( actual == '' )


class CueInputFixture(WebFixture, InputMixin2):
    cue_xpath = "//div[contains(@class,'reahl-cueinput')]/div[contains(@class,'reahl-cue')]/p"

    def new_text_input(self):
        return TextInput(self.form, self.field)

    def new_cue(self):
        return P(self.view, text='this is your cue')

    def new_cue_input(self):
        return CueInput(self.text_input, self.cue)

        
@test(CueInputFixture)
def cue_behaviour(fixture):
    """The Cue appears when its input has focus only."""
    model_object = fixture.model_object

    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(CueInput(TextInput(self, model_object.fields.an_attribute), P(view, text='this is your cue')))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    fixture.reahl_server.set_app(wsgi_app)
    fixture.driver_browser.open('/')

    # - initially rendered without cue
    vassert( fixture.driver_browser.is_element_present(fixture.cue_xpath) )
    fixture.driver_browser.wait_for_element_not_visible(fixture.cue_xpath)

    # - if click, cue appears
    fixture.driver_browser.click('//input[@type="text"]')
    fixture.driver_browser.wait_for_element_visible(fixture.cue_xpath)

    # - if focus lost, cue disappears again
    fixture.driver_browser.press_tab('//input')
    fixture.driver_browser.wait_for_element_not_visible(fixture.cue_xpath)


@test(CueInputFixture)
def basic_rendering_cue_input(fixture):
    """What the html for a CueInput looks like."""
    fixture.field.bind('an_attribute', fixture.model_object)
    cue = fixture.cue
    cue_input = fixture.cue_input
    fixture.form.add_child(cue_input)
    fixture.model_object.an_attribute = 'my value'

    tester = WidgetTester(cue_input)

    # Case: normal behaviour
    actual = tester.render_html()
    expected_html = '<div class="pure-g reahl-cueinput reahl-labelledinput">'\
                '<div class="column-label pure-u-1-4"><label for="an_attribute">the label</label></div>' \
                '<div class="column-input pure-u-1-2"><input name="an_attribute" form="test" type="text" value="my value" class="reahl-textinput"></div>' \
                 '<div class="column-cue pure-u-1-4 reahl-cue">' \
                 '<p hidden="true">this is your cue</p>' \
                 '</div>' \
               '</div>'
    vassert( actual == expected_html )

    # Case: when the underlying input is not visible
    fixture.field.access_rights.readable = Allowed(False)
    fixture.field.access_rights.writable = Allowed(False)
    actual = tester.render_html()
    vassert( actual == '' )


class FuzzyTextInputFixture(WebFixture, InputMixin2):
    def new_field(self, label='the label'):
        return DateField(label=label)


@test(FuzzyTextInputFixture)
def fuzzy(fixture):
    """A TextInput can be created as fuzzy=True. Doing this results in the possibly imprecise
       input that was typed by the user to be interpreted server-side and changed to the
       more exact representation in the client browser.  This happens upon the input losing focus."""
    model_object = fixture.model_object
    class MyForm(Form):
        def __init__(self, view, name):
            super(MyForm, self).__init__(view, name)
            self.add_child(LabelledBlockInput(TextInput(self, model_object.fields.an_attribute, fuzzy=True)))
            self.define_event_handler(model_object.events.an_event)
            self.add_child(Button(self, model_object.events.an_event))

    wsgi_app = fixture.new_wsgi_app(child_factory=MyForm.factory('myform'), enable_js=True)
    fixture.reahl_server.set_app(wsgi_app)
    browser = fixture.driver_browser
    browser.open('/')

    browser.type(XPath.input_labelled('the label'), '20 November 2012')
    browser.press_tab(XPath.input_labelled('the label'))
    browser.wait_for(browser.is_element_value, XPath.input_labelled('the label'), '20 Nov 2012')








# Copyright 2015 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Layout tools based on Pure (http://purecss.io/)

.. versionadded:: 3.1

"""
from __future__ import print_function, unicode_literals, absolute_import, division

from collections import OrderedDict
from fractions import Fraction

from reahl.web.fw import Layout

from reahl.web.ui import Panel, Header, Footer, Slot, HTML5Page
from reahl.component.exceptions import ProgrammerError, arg_checks, IsInstance


class UnitSize(object):
    """Represents a set of relative sizes to be used for an element depending on the size of the user's device.

       The Pure library sizes elements relative to their parent using
       fractions (1/2 or 3/4 or 1 are examples).  The size of an
       element relative to its parent can also vary, depending on the
       size of the device a user is using to view your web
       application.

       This class lets one specify the size to use for different sizes of
       display device.

       Sizes should be given as fractions in strings, such as the literal:

       .. code-block:: python

          UnitSize(sm='1', md='1/2', xl='1/4')

       Sizes can be given for sizes of device using the keyword arguments:

       :keyword default: If no other size is specified, or if the screen being used is smaller than the sizes given, this size is used.
       :keyword sm: For small screens (or larger).
       :keyword md: For medium-sized screen (or larger).
       :keyword lg: For large devices (or larger).
       :keyword xl: For extra-large devices (or larger).

       .. note::

          If a size is specified for `md`, for example, it will be used
          on all screens that are medium-sized or larger and Pure will fall
          back to using the `default` size for smaller screens.
    """
    def __init__(self, default=None, sm=None, md=None, lg=None, xl=None):
        self.default = default
        self.sm = sm
        self.md = md
        self.lg = lg
        self.xl = xl

    def as_dict(self):
        return dict(self.__dict__)


class ColumnLayout(Layout):
    """A Layout that divides an element into a number of columns.

       Each argument passed to the constructor defines a
       column. Columns are added to the element using this Layout in
       the order they are passed to the constructor. Columns can also 
       be added to the Widget later, by calling :meth:`ColumnLayout.add_column`.

       To define a column without specifying a size, just pass a
       string containing the column name.
       
       To define a column with a given UnitSize, pass a tuple of which
       the first element is the column name, and the second an
       instance of :class:`UnitSize`

       .. code-block:: python

          ColumnLayout('column_a', ('column_b', UnitSize(default='1/2')))

       .. admonition:: Styling
       
          Each column added is a <div> which has has the css class
          'column-<column_name>' where <column_name> is the name as
          specified to the constructor.

    """
    def __init__(self, *column_definitions):
        super(ColumnLayout, self).__init__()
        self.columns = OrderedDict()  #: A dictionary containing the added columns, keyed by column name.
        self.column_sizes = OrderedDict()
        for column_definition in column_definitions:
            if isinstance(column_definition, tuple):
                name, fractions = column_definition
            else:
                name, fractions = column_definition, UnitSize()
            self.column_sizes[name] = fractions

    def customise_widget(self):
        self.widget.append_class('pure-g')
        for name, unit_size in self.column_sizes.items():
             panel = self.add_column(unit_size=unit_size)
             self.columns[name] = panel
             panel.append_class('column-%s' % name)

    def add_column(self, unit_size=None):
        """Add an un-named column, optionally with the given :class:`UnitSize`.

           :keyword unit_size: The sizes to use for the column.

           Returns a :class:`reahl.web.ui.Panel` representing the added column.
        """
        unit_size = unit_size or UnitSize()
        unit = self.widget.add_child(Panel(self.view))

        for label, value in unit_size.as_dict().items():
            if label == 'default':
                unit.append_class(self.fraction_css_class('pure-u', unit_size.default) if unit_size.default else 'pure-u')
            elif value:
                unit.append_class(self.fraction_css_class('pure-u-%s' % label, value))

        return unit

    def fraction_css_class(self, prefix, fraction_string):
        fraction = Fraction(fraction_string)
        return '%s-%s-%s' % (prefix,  fraction.numerator, fraction.denominator)


class PageColumnLayout(Layout):
    """A Layout that provides the main layout for an :class:`reahl.web.ui.HTML5Page`.

       A PageColumnLayout adds a header and footer area to an HTML5Page, as well as 
       a number of columns between the header and footer area, as specified by the
       arguments to its constructor.
    
       All of these contents are also wrapped in a
       :class:`reahl.web.ui.Panel`, which is handy for styling.

       Specifying columns work exactly as for :class:`ColumnLayout`.

       Inside each added column, a :class:`reahl.web.ui.Slot` is added that is named
       after the column. 

       .. admonition:: Styling
       
          Adds a <div id="doc"> to the <body> of the page, which contains:

           - a <header id="hd">
           - a <div id="contents">
           - a <footer id="ft">

         The div#id element is further set up as with a :class:`ColumnLayout` using the
         arguments passed as `column_definition`\ s.
    """
    def __init__(self, *column_definitions):
        super(PageColumnLayout, self).__init__()
        self.document_layout = ColumnLayout(*column_definitions)
        self.header = None    #: The :class:`reahl.web.ui.Header` of the page.
        self.contents = None  #: The :class:`reahl.web.ui.Panel` containing the columns.
        self.footer = None    #: The :class:`reahl.web.ui.Footer` of the page.
        self.document = None  #: The :class:`reahl.web.ui.Panel` containing the entire page.

    def add_slots_for_columns(self):
        for column_name, column_widget in self.document_layout.columns.items():
            column_widget.add_child(Slot(self.view, column_name))

    @arg_checks(widget=IsInstance(HTML5Page))
    def apply_to_widget(self, widget):
        super(PageColumnLayout, self).apply_to_widget(widget)

    def customise_widget(self):
        self.document = self.widget.body.add_child(Panel(self.view))
        self.document.set_id('doc')
        
        self.header = self.document.add_child(Header(self.view))
        self.header.add_child(Slot(self.view, 'header'))
        self.header.set_id('hd')

        self.contents = self.document.add_child(Panel(self.view).use_layout(self.document_layout))
        self.add_slots_for_columns()

        self.contents.set_id('bd')
        self.contents.set_attribute('role', 'main')
        
        self.footer = self.document.add_child(Footer(self.view))
        self.footer.add_child(Slot(self.view, 'footer'))
        self.footer.set_id('ft')

    @property
    def columns(self):
        """A dictionary containing the added columns, keyed by column name."""
        return self.contents.layout.columns












# Copyright 2013, 2014 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert, expected
from reahl.stubble import EmptyStub

from reahl.web.ui import *
from reahl.webdev.tools import WidgetTester
from reahl.web_dev.fixtures import WebFixture

@test(WebFixture)
def basic_fixed_attributes(fixture):
    """How the static attributes of a Widget can be manipulated, queried and rendered."""

    widget = HTMLElement(fixture.view, 'x')
    tester = WidgetTester(widget)

    # Adding / setting
    widget.set_attribute('attr1', 'value1')
    widget.add_to_attribute('listattr', ['one', 'two'])
    widget.add_to_attribute('listattr', ['three'])
    
    # Querying
    vassert( widget.has_attribute('attr1') )
    vassert( widget.has_attribute('listattr') )
    vassert( not widget.has_attribute('notthere') )
    vassert( widget.attributes.v['attr1'] == 'value1' )
    vassert( widget.attributes.v['listattr'] == 'one three two' )
    
    # Rendering
    rendered = tester.render_html()
    vassert( rendered == '<x attr1="value1" listattr="one three two">' )

    # Rendering - the order of attributes in the rendered output
    widget.set_attribute('id', '123')
    widget.add_to_attribute('class', ['z', 'b'])
    rendered = tester.render_html()
    vassert( rendered == '<x id="123" attr1="value1" listattr="one three two" class="b z">' )


@test(WebFixture)
def handy_methods(fixture):
    """Some handy methods for special attributes."""

    widget = HTMLElement(fixture.view, 'x')
    tester = WidgetTester(widget)

    widget.set_title('the title')
    widget.set_id('the id')
    widget.append_class('two')
    widget.append_class('one')

    rendered = tester.render_html()
    vassert( rendered == '<x id="the id" title="the title" class="one two">' )

@test(WebFixture)
def dynamically_determining_attributes(fixture):
    """A Widget can determine its attribute values at the latest possible stage, based on changing data."""

    class WidgetWithDynamicAttributes(HTMLElement):
        state = '1'
        @property
        def attributes(self):
            attributes = super(WidgetWithDynamicAttributes, self).attributes
            attributes.set_to('dynamic', self.state)
            attributes.add_to('dynamiclist', [self.state])
            attributes.add_to('not-there', ['a value'])
            attributes.remove_from('not-there', ['a value'])
            return attributes

    widget = WidgetWithDynamicAttributes(fixture.view, 'x')
    widget.set_attribute('fixed', 'value1')
    tester = WidgetTester(widget)

    widget.state = '1'
    rendered = tester.render_html()
    vassert( rendered == '<x dynamic="1" dynamiclist="1" fixed="value1">' )

    widget.state = '2'
    rendered = tester.render_html()
    vassert( rendered == '<x dynamic="2" dynamiclist="2" fixed="value1">' )


@test(WebFixture)
def wrapper_widgets(fixture):
    """Sometimes, a HTMLElement is "wrapped" by an Input which represents the HTMLElement. In this case,
       dynamic as well as static attributes for the wrapped HTMLElement are obtained from its Input wrapper."""

    class WrapperWidget(Input):
        def get_wrapped_html_attributes(self, attributes):
            attributes.set_to('set-by-wrapper', 'rhythm and poetry')
            return attributes

    field = Field()
    field.bind('aname', field)
    wrapper = WrapperWidget(Form(fixture.view, 'formname'), field)
    widget = HTMLElement(fixture.view, 'x', wrapper_widget=wrapper)
    tester = WidgetTester(widget)

    # Case: dynamic attributes are supplied by the wrapper
    rendered = tester.render_html()
    vassert( rendered == '<x set-by-wrapper="rhythm and poetry">' )

    # Case:several methods to set static attributes are delegated from the wrapper
    wrapper.set_id('myid')
    wrapper.set_title('mytitle')
    wrapper.add_to_attribute('list-attribute', ['one', 'two'])
    wrapper.set_attribute('an-attribute', 'a value')

    rendered = tester.render_html()
    vassert( rendered == '<x id="myid" an-attribute="a value" list-attribute="one two" set-by-wrapper="rhythm and poetry" title="mytitle">' )

@test(WebFixture)
def all_html_widgets_have_css_ids(fixture):
    """A Widget (for HTML) can have a css_id. If accessed, but not set, a ProgrammerError is raised."""
    
    widget = HTMLElement(fixture.view, 'x')
    widget.tag_name = 'x'
    tester = WidgetTester(widget)

    widget._css_id = None
    with expected(ProgrammerError): 
        widget.css_id
    
    widget._css_id = 'myid'
    vassert( widget.css_id == 'myid' )

@test(WebFixture)
def jquery_support(fixture):
    """Each HTML Widget has a jquery_selector that targets it uniquely. By default
       this makes use of the css_id of the Widget, which thus need to be set, however
       subclasses may override this behaviour. The jquery_selector (or a more lax 
       selector for the Widget) can also be further narrowed to a specific
       jquery selector context."""
    
    widget = HTMLElement(fixture.view, 'x')
    tester = WidgetTester(widget)

    with expected(ProgrammerError):  # css id not set
        widget.jquery_selector

    widget.set_id('anid')
    vassert( widget.jquery_selector == '"#%s"' % widget.css_id )
    vassert( widget.attributes['id'].as_html_value() == widget.css_id )
    
    contextualised = widget.contextualise_selector('selector', 'context')
    vassert( contextualised ==  'selector, "context"' )

    contextualised = widget.contextualise_selector('selector', None)
    vassert( contextualised ==  'selector' )

@test(WebFixture)
def single_tags(fixture):
    """Definition of a HTMLElement."""
    
    single_tag = HTMLElement(fixture.view, 'single')
    tester = WidgetTester(single_tag)

    with expected(AssertionError):
        single_tag.add_child(P(fixture.view))

    rendered = tester.render_html()
    vassert( rendered == '<single>' )

@test(WebFixture)
def closing_tags(fixture):
    """Definition of a HTMLElement with children."""
    
    closing_tag = HTMLElement(fixture.view, 'closing', children_allowed=True)
    tester = WidgetTester(closing_tag)

    closing_tag.add_child(P(fixture.view))

    rendered = tester.render_html()
    vassert( rendered == '<closing><p></p></closing>' )


class Scenarios(WebFixture):
    @scenario
    def text_node(self):
        self.widget = TextNode(self.view, 'text')
        self.expected_html = 'text'

    @scenario
    def a1(self):
        self.widget = A(self.view, Url('/xyz'))
        self.expected_html = '<a href="/xyz"></a>'
        
    @scenario
    def a2(self):
        self.widget = A(self.view, Url('/xyz'), description='description')
        self.expected_html = '<a href="/xyz">description</a>'

    @scenario
    def a3(self):
        def disallowed(): return False
        self.widget = A(self.view, Url('/xyz'), description='description', write_check=disallowed)
        self.expected_html = '<a>description</a>'
        
    @scenario
    def h(self):
        self.widget = H(self.view, 2)
        self.expected_html = '<h2></h2>'
        
    @scenario
    def p1(self):
        self.widget = P(self.view)
        self.expected_html = '<p></p>'

    @scenario
    def p2(self):
        self.widget = P(self.view, text='text')
        self.expected_html = '<p>text</p>'

    @scenario
    def p_with_slots(self):
        template_p = P(self.view, text='the {0} {{brown}} {slot2} jumps')
        self.widget = template_p.format(Span(self.view, text='quick'), slot2=Span(self.view, text='fox'))
        self.expected_html = '<p>the <span>quick</span> {brown} <span>fox</span> jumps</p>'
        
    @scenario
    def title(self):
        self.widget = Title(self.view, 'text')
        self.expected_html = '<title>text</title>'
        
    @scenario
    def link(self):
        self.widget = Link(self.view, 'rr', 'hh', 'tt')
        self.expected_html = '<link href="hh" rel="rr" type="tt">'
        
    @scenario
    def header(self):
        self.widget = Header(self.view)
        self.expected_html = '<header></header>'
        
    @scenario
    def footer(self):
        self.widget = Footer(self.view)
        self.expected_html = '<footer></footer>'
        
    @scenario
    def li(self):
        self.widget = Li(self.view)
        self.expected_html = '<li></li>'
        
    @scenario
    def ul(self):
        self.widget = Ul(self.view)
        self.expected_html = '<ul></ul>'

    @scenario
    def img1(self):
        self.widget = Img(self.view, 'ss')
        self.expected_html = '<img src="ss">'
        
    @scenario
    def img2(self):
        self.widget = Img(self.view, 'ss', alt='aa')
        self.expected_html = '<img alt="aa" src="ss">'

    @scenario
    def span(self):
        self.widget = Span(self.view)
        self.expected_html = '<span></span>'
        
    @scenario
    def span_with_text(self):
        self.widget = Span(self.view, text='some text')
        self.expected_html = '<span>some text</span>'
        
    @scenario
    def div(self):
        self.widget = Div(self.view)
        self.expected_html = '<div></div>'

    @scenario
    def caption(self):
        self.widget = Caption(self.view, text='some text')
        self.expected_html = '<caption>some text</caption>'
        
    @scenario
    def col(self):
        self.widget = Col(self.view, span='2')
        self.expected_html = '<col span="2">'

    @scenario
    def colgroup(self):
        self.widget = Colgroup(self.view, span='2')
        self.expected_html = '<colgroup span="2"></colgroup>'

    @scenario
    def thead(self):
        self.widget = Thead(self.view)
        self.expected_html = '<thead></thead>'

    @scenario
    def tfoot(self):
        self.widget = Tfoot(self.view)
        self.expected_html = '<tfoot></tfoot>'

    @scenario
    def tbody(self):
        self.widget = Tbody(self.view)
        self.expected_html = '<tbody></tbody>'

    @scenario
    def tr(self):
        self.widget = Tr(self.view)
        self.expected_html = '<tr></tr>'

    @scenario
    def th(self):
        self.widget = Th(self.view)
        self.expected_html = '<th></th>'

    @scenario
    def td(self):
        self.widget = Td(self.view)
        self.expected_html = '<td></td>'

    @scenario
    def table(self):
        self.widget = Table(self.view, caption_text='my caption', summary='my summary')
        self.expected_html = '<table summary="my summary"><caption>my caption</caption></table>'
        
    @scenario
    def nav(self):
        self.widget = Nav(self.view)
        self.expected_html = '<nav></nav>'
        
    @scenario
    def article(self):
        self.widget = Article(self.view)
        self.expected_html = '<article></article>'
        
    @scenario
    def label1(self):
        self.widget = Label(self.view)
        self.expected_html = '<label></label>'
        
    @scenario
    def label2(self):
        self.widget = Label(self.view, text='text')
        self.expected_html = '<label>text</label>'
        
    @scenario
    def fieldset1(self):
        self.widget = FieldSet(self.view)
        self.expected_html = '<fieldset></fieldset>'
        
    @scenario
    def fieldset2(self):
        self.widget = FieldSet(self.view, label_text='text')
        self.expected_html = '<fieldset><label>text</label></fieldset>'

@test(Scenarios)
def basic_html_widgets(fixture):
    """Several basic widgets merely correspond to html elements."""
    
    tester = WidgetTester(fixture.widget)
    rendered_html = tester.render_html()
    vassert( rendered_html == fixture.expected_html )

@test(WebFixture)
def view_rights_propagate_to_a(fixture):
    """The access rights specified for a View are propagated to an A, made from a Bookmark to that View."""
    fixture.view.write_check = EmptyStub()
    fixture.view.read_check = EmptyStub()
    a = A.from_bookmark(fixture.view, fixture.view.as_bookmark())
    vassert( a.read_check is fixture.view.read_check )
    vassert( a.write_check is fixture.view.write_check )

@test(WebFixture)
def text_node_can_vary(fixture):
    """A TextNode can vary its text if constructed with a getter for the value instead of a hardcoded value."""
    def getter():
        return fixture.current_value
    widget = TextNode(fixture.view, getter)
    tester = WidgetTester(widget)
    
    fixture.current_value = 'stuff'
    rendered = tester.render_html()
    vassert( rendered == 'stuff' )

    fixture.current_value = 'other'
    rendered = tester.render_html()
    vassert( rendered == 'other' )

@test(WebFixture)
def text_node_escapes_html(fixture):
    """The text of a TextNode is html-escaped."""

    widget = TextNode(fixture.view, '<tag> "Here" & \'there\'')
    tester = WidgetTester(widget)
    
    rendered = tester.render_html()
    vassert( rendered == '&lt;tag&gt; "Here" &amp; \'there\'')


@test(WebFixture)
def literal_html(fixture):
    """The LiteralHTML Widget just renders a chunk of HTML, but can answer queries about images in that HTML."""
    
    contents = '<img src="_some_images/piet.pdf  "> <img src   = \' _some_images/koos was-^-hoêr.jpg\'>'
    literal_html = LiteralHTML(fixture.view, contents)
    tester = WidgetTester(literal_html)

    rendered_html = tester.render_html()
    vassert( rendered_html == contents )

    # Case: when the content is transformed
    def text_transformation(text):
        return text.replace('im', 'IM')
    literal_html = LiteralHTML(fixture.view, contents, transform=text_transformation)
    tester = WidgetTester(literal_html)

    rendered_html = tester.render_html()
    vassert( rendered_html == text_transformation(contents) )

@test(WebFixture)
def head(fixture):
    """Head corresponds with the head HTML element, can have a title and always has a special Slot used by the framework."""

    head = Head(fixture.view, 'a title')
    tester = WidgetTester(head)

    reahl_header_slot = head.children[1]
    vassert( isinstance(reahl_header_slot, Slot) )
    vassert( reahl_header_slot.name == 'reahl_header' )

    rendered_html = tester.render_html()
    vassert( rendered_html == '<head><title>a title</title></head>' )

@test(WebFixture)
def body(fixture):
    """Body corresponds with the body HTML element, and always has a special Slot at its end used by the framework."""
    body = Body(fixture.view)
    tester = WidgetTester(body)

    body.add_child(P(fixture.view))
    
    reahl_footer_slot = body.children[1]
    vassert( isinstance(reahl_footer_slot, Slot) )
    vassert( reahl_footer_slot.name == 'reahl_footer' )

    rendered_html = tester.render_html()
    vassert( rendered_html == '<body><p></p></body>' )




